!! Copyright (C) 2021 N. Tancogne-Dejean
!!
!! This program is free software; you can redistribute it and/or modify
!! it under the terms of the GNU General Public License as published by
!! the Free Software Foundation; either version 2, or (at your option)
!! any later version.
!!
!! This program is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU General Public License for more details.
!!
!! You should have received a copy of the GNU General Public License
!! along with this program; if not, write to the Free Software
!! Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
!! 02110-1301, USA.
!!
!!
! ---------------------------------------------------------
subroutine X(x_fbe_calc)(namespace, psolver, mesh, der, st, ex, vxc, fxc)
  type(namespace_t),        intent(in)    :: namespace
  type(poisson_t),          intent(in)    :: psolver
  class(mesh_t),            intent(in)    :: mesh
  type(derivatives_t),      intent(in)    :: der
  type(states_elec_t),      intent(inout) :: st
  real(real64),             intent(inout) :: ex
  real(real64), contiguous, optional, intent(inout) :: vxc(:,:) !< vxc(gr%mesh%np, st%d%nspin)
  real(real64), contiguous, optional, intent(out)   :: fxc(:,:,:) !< the exchange force density

  real(real64) :: eig
  integer :: isp
  real(real64), allocatable :: internal_vxc(:,:)

  PUSH_SUB_WITH_PROFILE(X(x_fbe_calc))

  ASSERT(st%d%ispin /= SPINORS)
  !At the moment we treat only spins and not k-points
  ASSERT(st%nik == st%d%spin_channels)

  if (present(vxc)) then
    SAFE_ALLOCATE(internal_vxc(1:mesh%np, 1:st%d%spin_channels))
    internal_vxc = M_ZERO
  end if
  if (present(fxc)) then
    fxc = M_ZERO
  end if

  eig = M_ZERO
  do isp = st%d%kpt%start, st%d%kpt%end
    if (present(vxc)) then
      call X(fbe_x)(namespace, mesh, der, psolver, st, isp, eig, vxc=internal_vxc(:,isp), fxc=fxc(:,:,isp))
    else
      call X(fbe_x)(namespace, mesh, der, psolver, st, isp, eig, fxc=fxc(:,:,isp))
    end if
  end do

  ! Reduction over spin channels
  if (st%d%kpt%parallel) then
    call comm_allreduce(st%d%kpt%mpi_grp, eig)
    if (present(vxc)) call comm_allreduce(st%d%kpt%mpi_grp, internal_vxc)
    if (present(fxc)) call comm_allreduce(st%d%kpt%mpi_grp, fxc)
  end if

  ex = ex + eig
  if (present(vxc)) then
    call lalg_axpy(mesh%np, st%d%spin_channels, M_ONE, internal_vxc, vxc)
  end if

  SAFE_DEALLOCATE_A(internal_vxc)

  POP_SUB_WITH_PROFILE(X(x_fbe_calc))
end subroutine X(x_fbe_calc)

!------------------------------------------------------------
!> This routine is adapted from the X(slater) routine
!------------------------------------------------------------
subroutine X(fbe_x) (namespace, mesh, der, psolver, st, isp, ex, vxc, fxc)
  type(namespace_t),           intent(in)    :: namespace
  class(mesh_t),               intent(in)    :: mesh
  type(derivatives_t),         intent(in)    :: der
  type(poisson_t),             intent(in)    :: psolver
  type(states_elec_t), target, intent(inout) :: st
  integer,                     intent(in)    :: isp
  real(real64),                intent(inout) :: ex
  real(real64), contiguous, optional, intent(inout) :: vxc(:) !< vxc(gr%mesh%np)
  real(real64), contiguous, optional, intent(inout) :: fxc(:,:) !< the exchange force density

  integer :: ii, jst, ist, nsteps, node_to, node_fr, ist_s, ist_r, idm, ip
  integer, allocatable :: recv_stack(:), send_stack(:)
  real(real64) :: rr, socc, tmp, ex_
  R_TYPE, allocatable :: rho_ij(:), pot_ij(:), psi(:,:), wf_ist(:,:), grad_rho_ij(:,:)
  real(real64), allocatable :: tmp_vxc(:,:), tmp_pot(:,:), div(:), grad_rho(:,:), slater(:)
  type(MPI_Request) :: send_req
  type(ring_pattern_t) :: ring_pattern

  PUSH_SUB(X(fbe_x))

  ASSERT(mesh%box%dim == 3) ! The present code is only in 3D, as otherwise the Poisson equation is different

  socc = M_ONE / st%smear%el_per_state

  SAFE_ALLOCATE(pot_ij(1:mesh%np))
  SAFE_ALLOCATE(grad_rho_ij(1:mesh%np, 1:mesh%box%dim))
  SAFE_ALLOCATE(rho_ij(1:mesh%np_part))
  SAFE_ALLOCATE(recv_stack(1:st%nst+1))
  SAFE_ALLOCATE(send_stack(1:st%nst+1))
  SAFE_ALLOCATE(psi(1:mesh%np, 1:st%d%dim))
  SAFE_ALLOCATE(wf_ist(1:mesh%np, 1:st%d%dim))

  if (present(vxc) .or. present(fxc)) then
    SAFE_ALLOCATE(tmp_pot(1:mesh%np, 1:mesh%box%dim))
    SAFE_ALLOCATE(tmp_vxc(1:mesh%np_part, 1:mesh%box%dim))
    tmp_vxc(:,:) = M_ZERO
    SAFE_ALLOCATE(slater(1:mesh%np_part))
    slater(:) = M_ZERO
    !We need the gradient of the density to evaluate the potential
    SAFE_ALLOCATE(grad_rho(1:mesh%np, 1:mesh%box%dim))
    call dderivatives_grad(der, st%rho(:, isp), grad_rho)
  end if

  ex_ = M_ZERO

  ! This is the maximum number of blocks for each processor
  call ring_pattern%start(st%mpi_grp, .true.)
  nsteps = ring_pattern%get_nsteps()

  do ii = 0, nsteps
    node_to = ring_pattern%get_rank_to(ii)
    node_fr = ring_pattern%get_rank_from(ii)

    ! check which wavefunctions we have to send/recv, and put them in a stack
    recv_stack(:) = -1
    ist_r = 1
    send_stack(:) = -1
    ist_s = 1

    do jst = 1, st%nst
      if (st%node(jst) == node_fr) then
        recv_stack(ist_r) = jst
        ist_r = ist_r + 1
      end if
      if (node_to /= -1 .and. st%node(jst) == st%mpi_grp%rank) then
        send_stack(ist_s) = jst
        ist_s = ist_s + 1
      end if
    end do

    ! now we do a loop over all states
    ist_r = 0
    if (node_fr < 0) ist_r = st%nst + 1
    ist_s = 0
    if (node_to < 0) ist_s = st%nst + 1
    do
      ! increment send counter
      if (ist_s <= st%nst) ist_s = ist_s  + 1

      if (st%parallel_in_states) then
        ! send wavefunction
        send_req = MPI_REQUEST_NULL
        if ((send_stack(ist_s) > 0).and.(node_to /= st%mpi_grp%rank)) then
          call states_elec_get_state(st, mesh, send_stack(ist_s), isp, psi)
          call st%mpi_grp%isend(psi, mesh%np*st%d%dim, R_MPITYPE, &
            node_to, send_req, tag=send_stack(ist_s))
        end if
      end if

      ! increment receive counter
      if (ist_r <= st%nst) ist_r = ist_r  + 1

      ! receive wavefunction
      if (recv_stack(ist_r) > 0) then
        if (node_fr == st%mpi_grp%rank) then
          call states_elec_get_state(st, mesh, send_stack(ist_r), isp, wf_ist)
        else
          if (st%parallel_in_states) then
            call st%mpi_grp%recv(wf_ist, mesh%np*st%d%dim, R_MPITYPE, &
              node_fr, recv_stack(ist_r))
          end if
        end if
      end if

      if (st%parallel_in_states) then
        if (send_req /= MPI_REQUEST_NULL) call st%mpi_grp%wait(send_req)
        send_req = MPI_REQUEST_NULL
      end if

      if (recv_stack(ist_r) > 0) then
        ! this is where we calculate the elements of the matrix
        ist = recv_stack(ist_r)
        do jst = st%st_start, st%st_end

          if (abs(st%occ(ist, isp)) < M_EPSILON .or. abs(st%occ(jst, isp)) < M_EPSILON) cycle

          if ((st%node(ist) == st%mpi_grp%rank).and.(jst < ist)) cycle
          if ((st%occ(ist, isp) <= M_EPSILON).or.(st%occ(jst, isp) <= M_EPSILON)) cycle

          call states_elec_get_state(st, mesh, jst, isp, psi)
          rho_ij(1:mesh%np) = R_CONJ(wf_ist(1:mesh%np, 1))*psi(1:mesh%np, 1)

          pot_ij(1:mesh%np) = R_TOTYPE(M_ZERO)
          call X(poisson_solve)(psolver, namespace, pot_ij, rho_ij, all_nodes=.false.)

          rr = socc * st%occ(ist, isp) * st%occ(jst, isp)
          if (ist /= jst) rr = rr * M_TWO

          !Here we accumulate the result for the potential
          if (present(vxc) .or. present(fxc)) then
            !There are two contributitions, the gradient of the Slater potential, and the
            !Poisson equation with the gradient of the exchange hole

            !We first compute the Slater contribution
            !$omp parallel do
            do ip = 1, mesh%np
              !If there is no density at this point, we simply ignore it
              slater(ip) = slater(ip) - rr * R_REAL(R_CONJ(rho_ij(ip)) * pot_ij(ip)) &
                / (SAFE_TOL(st%rho(ip, isp), M_EPSILON))
            end do

            call X(derivatives_grad)(der, rho_ij(:), grad_rho_ij(:, :))

            do idm = 1, mesh%box%dim
              !$omp parallel do private(tmp)
              do ip = 1, mesh%np
                !For numerical reasons, we need to compute the gradient of the \rho_ij/n this way
                !else, we have severe numerical troubles at the border of the simulation box
                !especially in the case of one electron only (\rho_ij = n)
                tmp = R_REAL(R_CONJ(grad_rho_ij(ip,idm)) * pot_ij(ip)) * st%rho(ip, isp) &
                  -R_REAL(R_CONJ(rho_ij(ip)) * pot_ij(ip)) * grad_rho(ip, idm)

                tmp_vxc(ip, idm) = tmp_vxc(ip, idm)  - rr * tmp &
                  / (SAFE_TOL(st%rho(ip, isp)**2,M_EPSILON**2))
              end do
            end do

          end if


          ! not an energy functional
          ! get the contribution to the exchange energy
          ex_ = ex_ - M_HALF * rr * R_REAL(X(mf_dotp)(mesh, rho_ij, pot_ij, reduce=.false.))
        end do
      end if

      ! all done?
      if ((send_stack(ist_s) < 0) .and. (recv_stack(ist_r) < 0)) exit
    end do
  end do

  if (st%parallel_in_states) then
    if (present(vxc) .or. present(fxc)) then
      call comm_allreduce(st%mpi_grp, tmp_vxc)
      call comm_allreduce(st%mpi_grp, slater)
    end if
    call comm_allreduce(st%mpi_grp, ex_)
  end if
  if (mesh%parallel_in_domains) call mesh%allreduce(ex_)
  ex = ex + ex_

  !First we add the part from the Slater potential
  if (present(vxc)) then
    call lalg_axpy(mesh%np, M_ONE, slater, vxc)
    SAFE_ALLOCATE(div(1:mesh%np_part))
    call dderivatives_div(der, tmp_vxc(:, :), div(:))
  end if

  ! We add the non-Slater part of the force density
  if(present(fxc)) then
    do idm = 1, mesh%box%dim
      !$omp parallel do
      do ip = 1, mesh%np
        fxc(ip, idm) = tmp_vxc(ip, idm)*st%rho(ip, isp)
      end do
    end do
  end if

  ! We add the non-Slater part to the potential
  if (present(vxc)) then
    tmp_vxc(:, 1) = M_ZERO
    call dpoisson_solve(psolver, namespace, tmp_vxc(:, 1), div(:), all_nodes=.false.)
    SAFE_DEALLOCATE_A(div)
    call lalg_axpy(mesh%np, M_ONE / M_PI / M_FOUR, tmp_vxc(:, 1), vxc)
  end if

  if(present(fxc)) then
    ! We compute the gradient of Slater and add it
    call dderivatives_grad(der, slater, tmp_vxc)
    do idm = 1, mesh%box%dim
      !$omp parallel do
      do ip = 1, mesh%np
        fxc(ip, idm) = -fxc(ip, idm) + tmp_vxc(ip, idm)*st%rho(ip, isp)
      end do
    end do
  end if
  SAFE_DEALLOCATE_A(tmp_vxc)

  SAFE_DEALLOCATE_A(recv_stack)
  SAFE_DEALLOCATE_A(send_stack)
  SAFE_DEALLOCATE_A(pot_ij)
  SAFE_DEALLOCATE_A(rho_ij)
  SAFE_DEALLOCATE_A(psi)
  SAFE_DEALLOCATE_A(wf_ist)
  SAFE_DEALLOCATE_A(slater)

  POP_SUB(X(fbe_x))
end subroutine X(fbe_x)


!! Local Variables:
!! mode: f90
!! coding: utf-8
!! End:
