!! Copyright (C) 2020 N. Tancogne-Dejean
!!
!! This program is free software; you can redistribute it and/or modify
!! it under the terms of the GNU General Public License as published by
!! the Free Software Foundation; either version 2, or (at your option)
!! any later version.
!!
!! This program is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU General Public License for more details.
!!
!! You should have received a copy of the GNU General Public License
!! along with this program; if not, write to the Free Software
!! Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
!! 02110-1301, USA.
!!

#include "global.h"

!> @brief This module defines the quantity_t class and the IDs for quantities, which can be exposed by a system,
!! and used by an interaction.
!!
!! A special case of quantities, the so-called couplings are used to transfer information
!! from a system to the interactions. Other quantities can be used for output purposes.
!!
!! An interaction partner (derived from interaction_partner_t) needs to update the behaviour of the quantity by setting
!!
!! this%quantities(<quantity_id>)%updated_on_demand
!! this%quantities(<quantity_id>)%required
!! this%quantities(<quantity_id>)%always_available
!!
!! For the details of these flags see the documentation of quantity_t
!!
!! Interactions also need to declare, which quantities they require from the system and the interaction partner:
!!
!!  ! From the system:
!!  this%system_quantities = [<system_quantity1_id>, <system_quantity2_id>, ...]
!!
!!  ! From the partner:
!!  this%couplings_from_partner = [<partner_quantity1_id>, <partner_quantity1_id>, ...]
!!

module quantity_oct_m
  use iteration_counter_oct_m
  implicit none

  private
  public ::                   &
    quantity_t

  !# doc_start quantity
  integer, public, parameter ::         &
    POSITION                     =  1,  &
    VELOCITY                     =  2,  &
    CURRENT                      =  3,  &
    DENSITY                      =  4,  &
    SCALAR_POTENTIAL             =  5,  &
    VECTOR_POTENTIAL             =  6,  &
    E_FIELD                      =  7,  &
    B_FIELD                      =  8,  &
    MASS                         =  9,  &
    CHARGE                       = 10,  &
    PERMITTIVITY                 = 11,  &
    PERMEABILITY                 = 12,  &
    E_CONDUCTIVITY               = 13,  &
    M_CONDUCTIVITY               = 14,  &
    DIPOLE                       = 15,  &
    MAX_QUANTITIES               = 15
  !# doc_end

  character(len=17), public, parameter :: QUANTITY_LABEL(MAX_QUANTITIES) = (/ &
    "position        ", &
    "velocity        ", &
    "current         ", &
    "density         ", &
    "scalar potential", &
    "vector potential", &
    "E field         ", &
    "B field         ", &
    "mass            ", &
    "charge          ", &
    "permittivity    ", &
    "permeability    ", &
    "e_conductivity  ", &
    "m_conductivity  ", &
    "dipole          "  &
    /)

  !> Systems (system_t) can expose quantities that can be used to calculate interactions
  !! with other systems.
  !!
  !! Some quantities are dynamical variables of the system. Such quantities are
  !! usually updated by the propagation algorithm and cannot be calculated
  !! on-demand.
  type quantity_t
    private
    type(iteration_counter_t), public :: iteration      !< @brief Iteration counter storing the time at which the quantity was last updated.
    !                                                   !!
    !                                                   !! @note quantities dont have a constructor. The iteration counter is initialized automatically
    !                                                   !! for extenstions of system_t, but needs to be initialized explicitly for extensions of
    !                                                   !! interaction_partner_t, which are not systems.

    logical,       public :: required = .false.         !< @brief Should this quantities be calculated?
    !                                                   !!
    !                                                   !! true means that they are needed by either some interaction
    !                                                   !! or because they are dynamical variables of the system
    !                                                   !! (i.e., they appear in the differential equations governing
    !                                                   !! the systems dynamics). This is just a flag saying that the
    !                                                   !! system needs to compute and store these quantities. It says
    !                                                   !! nothing about **when** they should be updated.

    logical,       public :: always_available = .false. !< @brief Can we use this quantity at any requested iteration?
    !                                                   !!
    !                                                   !! (e.g., this will be true for a static quantity,
    !                                                   !! but false for a quantity that is only updated at specific iterations)

    logical,       public :: updated_on_demand = .true. !< @brief when is the quantity updated?
    !                                                   !!
    !                                                   !! If true, the quantity is only updated when requested. The quantity should be calculated (updated)
    !                                                   !! in the systems update_quantity() routine.
    !                                                   !!
    !                                                   !! If false, the quantity is updated automatically during the execution of an algorithm.
    !                                                   !! In this case, the quantity should be updated inside the do_algorithmic_operation() routine.

  end type quantity_t

contains

end module quantity_oct_m

!! Local Variables:
!! mode: f90
!! coding: utf-8
!! End:
